
assert( LibStub, "LibStub is not loaded." )

local addon = LibStub( "AceAddon-3.0" ):NewAddon( "hcRepairBot", "AceConsole-3.0", "AceEvent-3.0" )

local DEFAULTS = {
  profile = {
		
		-- True if player-funded repairs should be permitted in any 
		-- situation.
		playerEnabled = true,
		-- The minimum faction standing with the merchant to permit 
		-- player-funded repairs.
		playerReputation = 0,
		-- The minimum amount of money the player should reserve. 
		playerReserve = "0",

		-- True if guild-funded repairs should be permitted in any 
		-- situation.
		guildEnabled = true,
		-- The minimum faction standing with the merchant to permit 
		-- guild-funded repairs.
		guildReputation = 0,
		
		-- True if guild-funded repairs should be permitted when the 
		-- player is in a raid.				
		raidEnabled = true,
		-- True if guild-funded repairs should be permitted only when the 
		-- player is in a guild raid.
		raidGuild = true,

		-- True if guild-funded repairs should be permitted when the 
		-- player is in any party.
		partyEnabled = false, 
		-- True if guild-funded repairs should be permitted only when the 
		-- player is in a guild party.
		partyGuild = false,
		
		-- True if guild-funded repairs should be permitted when the
		-- player is not in a group.
		soloEnabled = false,
		
  },
}

local function _IsPlayerRepairDisabled( info )
	return not addon.db.profile.playerEnabled
end

local function _IsGuildRepairDisabled( info )
	return not addon.db.profile.guildEnabled
end

local function _IsPartyRepairDisabled( info )
	return not addon.db.profile.partyEnabled or _IsGuildRepairDisabled( info )
end

local function _IsRaidRepairDisabled( info )
	return not addon.db.profile.raidEnabled or _IsGuildRepairDisabled( info )
end

local function _GetAddonOption( info )
	return addon.db.profile[ info[ #info ] ]
end

local function _SetAddonOption( info, value )
	addon.db.profile[ info[ #info ] ] = value
end

local OPTIONS_REPUTATION_VALUES = {
	[ 0 ] = "None",
	[ 5 ] = "Friendly",
	[ 6 ] = "Honored",
	[ 7 ] = "Revered",
	[ 8 ] = "Exalted",
}

local OPTIONS = {
	type = "group",
	name = "Repair",
	get = _GetAddonOption,
	set = _SetAddonOption,
	args = {
		player = {
			type = "group",
			order = 10,
			name = "Player",
			inline = true,
			args = {
				playerEnabled = {	
					type = "toggle",
					order = 1,
					width = "full",
					name = "Repair Using Player Funds",
					desc = "Enable or disable player-funded automatic repairs.",
					descStyle = "inline",
					disabled = false,
				},
				playerReputation = {
					type = "select",
					order = 2,
					name = "Reputation Threshold",
					desc = "The minimum reputation required with the merchant to perform player-funded automatic repairs.",
					values = OPTIONS_REPUTATION_VALUES,
					disabled = _IsPlayerRepairDisabled,
				},
				reservePrefix = { -- this is just to add whitespace in the interface options frame
					type = "description",
					order = 3,
					width = "full",
					name = "",
				},
				playerReserve = {
					type = "input",
					order = 4,
					name = "Gold Reserve",
					pattern = "^[%d]+$",
					usage = "This must be a numeric value.",
					disabled = _IsPlayerRepairDisabled,
				},
				reserveSuffix = {
					type = "description",
					order = 5,
					width = "full",
					name = "The minimum quantity of gold to reserve. Automatic repairs will not be performed if the character does not have at least this much gold.",
				},
			},
		},
		guild = {
			type = "group",
			order = 20,
			name = "Guild",
			inline = true,
			args = {
				guildEnabled = {
					type = "toggle",
					order = 1,
					width = "full",
					name = "Repair Using Guild Funds",
					desc = "Enable guild-funded automatic repairs. If not in a guild which pays for repairs, these options have no effect.",
					descStyle = "inline",
					disabled = false,
				},
				guildReputation = {
					type = "select",
					order = 2,
					name = "Reputation Threshold",
					desc = "The minimum reputation required with the merchant to perform guild-funded automatic repairs.",
					values = OPTIONS_REPUTATION_VALUES,
					disabled = _IsGuildRepairDisabled,
				},
				raidHeader = {
					type = "header",
					order = 10,
					name = "Raid",
				},
				raidEnabled = {	
					type = "toggle",
					order = 11,
					width = "full",
					name = "Repair while in a Raid",
					desc = "Enable guild-funded automatic repairs while in a raid.",
					descStyle = "inline",
					disabled = _IsGuildRepairDisabled,
				},
				raidGuild = {	
					type = "toggle",
					order = 12,
					width = "full",
					name = "Guild-Only",
					desc = "Enable guild-funded automatic repairs only while in a guild raid.",
					descStyle = "inline",
					disabled = _IsRaidRepairDisabled,
				},
				partyHeader = {
					type = "header",
					order = 20,
					name = "Party",
				},
				partyEnabled = {	
					type = "toggle",
					order = 21,
					width = "full",
					name = "Repair while in a Party",
					desc = "Enable guild-funded automatic repairs while in a party.",
					descStyle = "inline",
					disabled = _IsGuildRepairDisabled
				},
				partyGuild = {	
					type = "toggle",
					order = 22,
					width = "full",
					name = "Guild-Only",
					desc = "Enable guild-funded automatic repairs only while in a guild party.",
					descStyle = "inline",
					disabled = _IsPartyRepairDisabled,
				},
				soloHeader = {
					type = "header",
					order = 30,
					name = "Solo",
				},
				soloEnabled = {	
					type = "toggle",
					order = 31,
					width = "full",
					name = "Repair while Solo",
					desc = "Enable guild-funded automatic repairs while not in a party or raid.",
					descStyle = "inline",
					disabled = _IsGuildRepairDisabled,
				},
			},
		},
	},
}

local function _SetDefaultOptions( )
	addon.db:ResetProfile( )
	LibStub( "AceConfigRegistry-3.0" ):NotifyChange( addon.name )
end

local function _CreateAddonOptionFrame( addon, options )
	local name = addon.name
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( name, options )
	local frame = LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( name, name )
	frame.default = _SetDefaultOptions
end

local function _CreateProfileOptionFrame( addon )
	local label = "Profiles"
	local name = addon.name .. label
	addon.profiles = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( addon.db )
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( name, addon.profiles )
	LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( name, label, addon.name )
end

--- Prepares the addon options, internal data, and configuration
function addon:OnInitialize( )
	self.db = LibStub( "AceDB-3.0" ):New( "hcRepairBotSavedVariables", DEFAULTS, true )
	self.state = {
		cost = 0, -- total cost of repairs
		guildBudget = 0, -- guild money available for repairs
		guildSpent = 0, -- guild money spent on repairs
		playerBudget = 0, -- player money available for repairs
		playerSpent = 0, -- player money spent on repairs
	}
	self.tooltip = CreateFrame( "GameTooltip", self.name .. "Tooltip", UIParent, "GameTooltipTemplate" )

	_CreateAddonOptionFrame( self, OPTIONS )
	_CreateProfileOptionFrame( self )
end

--- Registers merchant interaction events.
function addon:OnEnable( )
	self:RegisterEvent( "MERCHANT_CLOSED" )
	self:RegisterEvent( "MERCHANT_SHOW" )
end

-- Tests whether or not the player meets the reputation criteria 
-- set for automatic repairs.
-- @param profile repair options
-- @return true if the player meets the reputation criteria.
local function _IsPlayerReputable( standing )
	return ( UnitReaction( "npc", "player" ) >= standing )
end

-- @return true if the player is in a raid group, false otherwise.
local function _IsPlayerInRaid( )
	return GetNumRaidMembers( ) > 0
end

-- @return the current raid size, as defined by the raid difficulty.
local function _GetRaidDifficultySize( )
	local difficulty = GetCurrentRaidDifficulty( )
	if ( difficulty == 1 or difficulty == 3 ) then
		return 10
	else
		return 25
	end
end

-- @return true if the player is in a raid group with eight or twenty 
-- guild members, for ten-and twenty-five-man raids, repsectively.
local function _IsPlayerInGuildRaid( )
	local threshold = _GetRaidDifficultySize( ) * 0.8
	local numGuildMembers = 0
	
	for raidId = 1, GetNumRaidMembers( ) do
		if UnitIsInMyGuild( "raid" .. raidId ) then
			numGuildMembers = numGuildMembers + 1
		end
	end
	
	return ( numGuildMembers >= threshold )
end

-- @return true if the player is in a party.
local function _IsPlayerInParty( )
	return GetNumPartyMembers( ) > 0
end

-- @return true if the player is in a party with four or more guild members.
local function _IsPlayerInGuildParty( )
	local threshold = 4
	for partyId = 1, numPartyMembers do
		numGuildMembers = numGuildMembers + 1
	end
	return ( numGuildMembers >= threshold )
end

--- Gets the amount of guild money available for repairs.
-- If the player is not in a guild, or does not have access to guild 
-- repair funds, or has exceeded the daily guild repair limit, or 
-- does not pass the other criteria set in this addon's configuration 
-- options, this will return zero. When money is available, the 
-- amount is returned.
-- @return the guild repair budget, in copper.
function addon:GetGuildRepairBudget( )
	if not IsInGuild( ) then
		return 0
	end
	
	if not CanGuildBankRepair( ) then
		return 0
	end
	
	if _IsGuildRepairDisabled( ) then
		return 0
	end
	
	if not _IsPlayerReputable( self.db.profile.guildReputation ) then
		return 0
	end

	if _IsPlayerInRaid( ) then
		if not self.db.profile.raidEnabled then
			return 0
		end
		if self.db.profile.raidGuild and not _IsPlayerInGuildRaid( ) then
			return 0
		end
	elseif _IsPlayerInParty( ) then
		if not self.db.profile.partyEnabled then
			return 0
		elseif self.db.profile.partyGuild and _IsPlayerInGuildParty( ) then
			return 0
		end
	elseif not self.db.profile.soloEnabled then
		return 0
	end
	
	return GetGuildBankWithdrawMoney( )
end

--- Gets amount of player money available for repairs.
-- This returns the amount of copper available for repairs, subject 
-- to the reserve amount set in this addon's options. If player 
-- repairs are not enabled, or the merchant does not meet the 
-- player's reputation requirement, zero is returned.
-- @return an amount of copper available for repairs.
function addon:GetPlayerRepairBudget( )
	if _IsPlayerRepairDisabled( ) then
		return 0
	end
	
	if not _IsPlayerReputable( self.db.profile.playerReputation ) then
		return 0
	end

	local reserve = tonumber( self.db.profile.playerReserve ) * 10000
	return math.max( GetMoney( ) - reserve, 0 )
end

-- Calculates the repair cost of an item in the player's bags.
local function _GetContainerItemRepairCost( container, slot )
	if not InRepairMode( ) then
		ShowRepairCursor( )
	end
	
	local hasCooldown, repairCost = addon.tooltip:SetBagItem( container, slot )
	
	return repairCost or 0
end

-- Calculates the repair cost of an item in the player's inventory.
local function _GetInventoryItemRepairCost( slot )
	if not InRepairMode( ) then
		ShowRepairCursor( )
	end
	
	local hasItem, _, repairCost = addon.tooltip:SetInventoryItem( "player", slot )
	if hasItem then
		return repairCost
	end
	
	return 0
end

--- Repairs all items using guild or player funds.
-- This should only be called when the repair budget allows for a full 
-- repair. 
-- @param guild true if guild money should be used, false for player 
-- money.
function addon:RepairAllItems( guild )
	if guild then
		assert( self.state.guildBudget >= self.state.cost )
		RepairAllItems( 1 )
		self.state.guildSpent = self.state.guildSpent + self.state.cost
		self.state.guildBudget = self.state.guildBudget - self.state.cost
	else
		assert( self.state.playerBudget >= self.state.cost )
		RepairAllItems( )
		self.state.playerSpent = self.state.playerSpent + self.state.cost
		self.state.playerBudget = self.state.playerBudget - self.state.cost
	end
	
	self.state.cost = 0
end

-- @return true if the player prefers ranged over melee weapons.
local function _IsPlayerRanged( )
	local _, class = UnitClass( "player" )
	return class == "HUNTER"
end

--- Repairs a single item in the player's inventory.
-- Single-item repairs are only possible using player funds. 
-- @param slot an inventory slot.
function addon:RepairInventoryItem( slot )
	local link = GetInventoryItemLink( "player", slot )
	if not link then
		return
	end
	
	local cost = _GetInventoryItemRepairCost( slot )
	if ( cost > self.state.playerBudget ) or ( cost <= 0 ) then
		return
	end
	
	if not InRepairMode( ) then
		ShowRepairCursor( )
	end
	PickupInventoryItem( slot )
	
	self.state.playerSpent = self.state.playerSpent + cost
	self.state.playerBudget = self.state.playerBudget - cost
	self.state.cost = self.state.cost - cost
end

--- Repairs individual items in the player's inventory.
-- This will repair items until the total remaining cost of repairs 
-- is less than the guild repair budget, allowing guild repair funds 
-- to be used as much as possible. This walks through inventory 
-- slots in the internal game order. Called without any arguments, 
-- this will first repair they player's weapons, the the rest of 
-- the player's gear.
-- @param first the first inventory slot to repair.
-- @param last the last inventory slot to repair.
function addon:RepairSomeItems( first, last )

	if not first then
		if _IsPlayerRanged( ) then
			self:RepairSomeItems( INVSLOT_RANGED )
		else
			self:RepairSomeItems( INVSLOT_MAINHAND, INVSLOT_OFFHAND )
		end
		self:RepairSomeItems( INVSLOT_FIRST_EQUIPPED, INVSLOT_LAST_EQUIPPED )
		return
	end
	
	if not last then
		last = first
	end
	
	for slot = first, last do
		self:RepairInventoryItem( slot )
		if self.state.cost <= self.state.guildBudget then
			break
		end
	end
end

--- Repairs items using guild and player money.
-- If the guild's repair budget is insufficient to cover the full cost 
-- of repairs, player money will be used to repair some items until 
-- the guild budget can be used to cover the full cost of repairs. If 
-- the player budget is inadequate to do this, player money will 
-- continue to be spent until all items are repaired or the player 
-- budget is exhausted.
-- @param cost the total cost of repairs,
-- @param guildBudget the guild money available for repairs.
-- @param playerBudget the player money available for repairs.
-- @return the remaining repair cost.
function addon:RepairUsingAllMoney( )
	if self.state.cost > self.state.guildBudget then
		self:RepairSomeItems( )
	end
	self:RepairUsingGuildMoney( )
	self:RepairUsingPlayerMoney( )
end

--- Repairs items using guild money.
-- If the guild's repair budget is insufficient to cover the full cost 
-- of repairs, no action will be taken.
-- @param cost the total cost of repairs.
-- @param budget the repair budget.
-- @return the remaining cost of repairs.
function addon:RepairUsingGuildMoney( )
	if self.state.cost <= self.state.guildBudget then
		self:RepairAllItems( true )
	end
end

--- Repairs items using player money. 
-- If the player's repair budget is insufficient to cover the full cost 
-- of repairs, this will instead repair only what items can fit within 
-- that budget.
-- @param cost the total cost of repairs.
-- @param budget the repair budget.
-- @return the remaining cost of repairs.
function addon:RepairUsingPlayerMoney( )
	if self.state.cost <= self.state.playerBudget then
		self:RepairAllItems( false )
	else
		self:RepairSomeItems( )
	end
end

--- Clears the internal state table.
-- Sets the repair cost, budgets, and expenditures to zero.
function addon:ClearState( )
	self.state.cost = 0
	self.state.guildBudget = 0
	self.state.guildSpent = 0
	self.state.playerBudget = 0
	self.state.playerSpent = 0
end

--- Updates the internal state table.
-- This resets the repair cost, budgets, and expenditures when 
-- interacting with a new merchant.
function addon:UpdateState( )
	self.state.cost = GetRepairAllCost( )
	self.state.guildBudget = self:GetGuildRepairBudget( )
	self.state.guildSpent = 0
	self.state.playerBudget = self:GetPlayerRepairBudget( )
	self.state.playerSpent = 0
end

--- Fires when the player ends interaction with a merchant.
-- This clears the internal state table.
function addon:MERCHANT_CLOSED( )
	self:ClearState( )
end

--- Fires when the player begins interaction with a merchant.
-- This will determine whether or not the merchant is capable of 
-- performing repairs. If so, guild and/or player money will be used 
-- to perform such repairs, subject to the addon configuratoin 
-- options.
function addon:MERCHANT_SHOW( )
	if not CanMerchantRepair( ) then
		return
	end

	self:UpdateState( )
	
	if self.state.cost <= 0 then
		return
	end

	if self.state.guildBudget > 0 then
		if self.state.playerBudget > 0 then
			self:RepairUsingAllMoney( )
		else
			self:RepairUsingGuildMoney( )
		end
	else
		if self.state.playerBudget > 0 then
			self:RepairUsingPlayerMoney( )
		end
	end

	if self.state.guildSpent > 0 then
		local spent = GetCoinTextureString( self.state.guildSpent )		
		ChatFrame_DisplaySystemMessageInPrimary( ( "Your guild spent %s on repairs." ):format( spent ) )
	end

	if self.state.playerSpent > 0 then
		local spent = GetCoinTextureString( self.state.playerSpent )
		ChatFrame_DisplaySystemMessageInPrimary( ( "You spent %s on repairs." ):format( spent ) )
	end
	
	ResetCursor( )
end
